@@ -2,7 +2,7 @@
  * VLCPlaybackController.m
  * VLC for iOS
  *****************************************************************************
- * Copyright (c) 2013-2015 VideoLAN. All rights reserved.
+ * Copyright (c) 2013-2018 VideoLAN. All rights reserved.
  * $Id$
  *
  * Authors: Felix Paul Kühne <fkuehne # videolan.org>
@@ -54,6 +54,7 @@ typedef NS_ENUM(NSUInteger, VLCAspectRatio) {
     NSTimer *_sleepTimer;
 
     NSUInteger _currentAspectRatio;
+    BOOL _isInFillToScreen;
 
     UIView *_videoOutputViewWrapper;
     UIView *_actualVideoOutputView;
@@ -62,13 +63,12 @@ typedef NS_ENUM(NSUInteger, VLCAspectRatio) {
     BOOL _needsMetadataUpdate;
     BOOL _mediaWasJustStarted;
     BOOL _recheckForExistingThumbnail;
-    BOOL _headphonesWasPlugged;
+    BOOL _externalAudioPlaybackDevicePlugged;
 
     NSLock *_playbackSessionManagementLock;
 
-    VLCDialogProvider *_dialogProvider;
-
     NSMutableArray *_shuffleStack;
+    void (^_playbackCompletion)(BOOL success);
 }
 
 @end
@@ -100,7 +100,7 @@ typedef NS_ENUM(NSUInteger, VLCAspectRatio) {
     self = [super init];
     if (self) {
         // listen to audiosessions and appkit callback
-        _headphonesWasPlugged = [self areHeadphonesPlugged];
+        _externalAudioPlaybackDevicePlugged = [self isExternalAudioPlaybackDevicePlugged];
         NSNotificationCenter *defaultCenter = [NSNotificationCenter defaultCenter];
         [defaultCenter addObserver:self selector:@selector(audioSessionRouteChange:)
                               name:AVAudioSessionRouteChangeNotification object:nil];
@@ -117,7 +117,8 @@ typedef NS_ENUM(NSUInteger, VLCAspectRatio) {
                               name:UIApplicationDidEnterBackgroundNotification object:nil];
 
         _metadata = [VLCMetaData new];
-        _dialogProvider = [[VLCDialogProvider alloc] initWithLibrary:[VLCLibrary sharedLibrary] customUI:NO];
+        _dialogProvider = [[VLCDialogProvider alloc] initWithLibrary:[VLCLibrary sharedLibrary] customUI:YES];
+        _dialogProvider.customRenderer = self;
 
         _playbackSessionManagementLock = [[NSLock alloc] init];
         _shuffleMode = NO;
@@ -141,8 +142,14 @@ typedef NS_ENUM(NSUInteger, VLCAspectRatio) {
     [_mediaPlayer addPlaybackSlave:[NSURL fileURLWithPath:pathToFile] type:VLCMediaPlaybackSlaveTypeSubtitle enforce:YES];
 }
 
-- (void)playMediaList:(VLCMediaList *)mediaList firstIndex:(NSInteger)index subtitlesFilePath:(NSString *)subsFilePath
+- (void)playMediaList:(VLCMediaList *)mediaList firstIndex:(NSInteger)index subtitlesFilePath:(NSString * _Nullable)subsFilePath
 {
+    [self playMediaList: mediaList firstIndex: index subtitlesFilePath: subsFilePath completion: nil];
+}
+
+- (void)playMediaList:(VLCMediaList *)mediaList firstIndex:(NSInteger)index subtitlesFilePath:(NSString * _Nullable)subsFilePath completion:(void (^ __nullable)(BOOL success))completion
+{
+    _playbackCompletion = completion;
     self.mediaList = mediaList;
     _itemInMediaListToBePlayedFirst = (int)index;
     _pathToExternalSubtitlesFile = subsFilePath;
@@ -185,10 +192,7 @@ typedef NS_ENUM(NSUInteger, VLCAspectRatio) {
     _actualVideoOutputView.autoresizingMask = UIViewAutoresizingFlexibleWidth | UIViewAutoresizingFlexibleHeight;
     _actualVideoOutputView.autoresizesSubviews = YES;
 
-    if (_pathToExternalSubtitlesFile)
-        _listPlayer = [[VLCMediaListPlayer alloc] initWithOptions:@[[NSString stringWithFormat:@"--%@=%@", kVLCSettingSubtitlesFilePath, _pathToExternalSubtitlesFile]] andDrawable:_actualVideoOutputView];
-    else
-        _listPlayer = [[VLCMediaListPlayer alloc] initWithDrawable:_actualVideoOutputView];
+    _listPlayer = [[VLCMediaListPlayer alloc] initWithDrawable:_actualVideoOutputView];
 
     /* to enable debug logging for the playback library instance, switch the boolean below
      * note that the library instance used for playback may not necessarily match the instance
@@ -199,12 +203,8 @@ typedef NS_ENUM(NSUInteger, VLCAspectRatio) {
     [_mediaPlayer setDelegate:self];
     if ([[defaults objectForKey:kVLCSettingPlaybackSpeedDefaultValue] floatValue] != 0)
         [_mediaPlayer setRate: [[defaults objectForKey:kVLCSettingPlaybackSpeedDefaultValue] floatValue]];
-    if ([[defaults objectForKey:kVLCSettingDeinterlace] intValue] != 0)
-        [_mediaPlayer setDeinterlaceFilter:@"blend"];
-    else
-        [_mediaPlayer setDeinterlaceFilter:nil];
-    if (_pathToExternalSubtitlesFile)
-        [_mediaPlayer addPlaybackSlave:[NSURL fileURLWithPath:_pathToExternalSubtitlesFile] type:VLCMediaPlaybackSlaveTypeSubtitle enforce:YES];
+    int deinterlace = [[defaults objectForKey:kVLCSettingDeinterlace] intValue];
+    [_mediaPlayer setDeinterlace:deinterlace withFilter:@"blend"];
 
     VLCMedia *media = [_mediaList mediaAtIndex:_itemInMediaListToBePlayedFirst];
     [media parseWithOptions:VLCMediaParseLocal];
@@ -238,6 +238,8 @@ typedef NS_ENUM(NSUInteger, VLCAspectRatio) {
     [_mediaPlayer addObserver:self forKeyPath:@"time" options:0 context:nil];
     [_mediaPlayer addObserver:self forKeyPath:@"remainingTime" options:0 context:nil];
 
+    [_mediaPlayer setRendererItem:_renderer];
+
     [_listPlayer playItemAtNumber:@(_itemInMediaListToBePlayedFirst)];
 
     if ([self.delegate respondsToSelector:@selector(prepareForMediaPlayback:)])
@@ -249,6 +251,17 @@ typedef NS_ENUM(NSUInteger, VLCAspectRatio) {
 
     [[self remoteControlService] subscribeToRemoteCommands];
 
+    if (_pathToExternalSubtitlesFile) {
+        /* this could be a path or an absolute string - let's see */
+        NSURL *subtitleURL = [NSURL URLWithString:_pathToExternalSubtitlesFile];
+        if (!subtitleURL || !subtitleURL.scheme) {
+            subtitleURL = [NSURL fileURLWithPath:_pathToExternalSubtitlesFile];
+        }
+        if (subtitleURL) {
+            [_mediaPlayer addPlaybackSlave:subtitleURL type:VLCMediaPlaybackSlaveTypeSubtitle enforce:YES];
+        }
+    }
+
     _playerIsSetup = YES;
 
     [[NSNotificationCenter defaultCenter] postNotificationName:VLCPlaybackControllerPlaybackDidStart object:self];
@@ -284,30 +297,13 @@ typedef NS_ENUM(NSUInteger, VLCAspectRatio) {
     }
     if (!_sessionWillRestart) {
         _mediaList = nil;
-        if (_pathToExternalSubtitlesFile) {
-            NSFileManager *fileManager = [NSFileManager defaultManager];
-            if ([fileManager fileExistsAtPath:_pathToExternalSubtitlesFile])
-                [fileManager removeItemAtPath:_pathToExternalSubtitlesFile error:nil];
-            _pathToExternalSubtitlesFile = nil;
-        }
     }
     _playerIsSetup = NO;
     [_shuffleStack removeAllObjects];
 
-    if (@available(iOS 10, *)) {
-        if (_errorCallback && _mediaPlayer.state == VLCMediaPlayerStateError &&  !_sessionWillRestart)
-            [[UIApplication sharedApplication] openURL:_errorCallback options:@{} completionHandler:nil];
-        else if (_successCallback && !_sessionWillRestart)
-            [[UIApplication sharedApplication] openURL:_successCallback options:@{} completionHandler:nil];
-    } else {
-#pragma clang diagnostic push
-#pragma clang diagnostic ignored "-Wdeprecated-declarations"
-        /* UIApplication's replacement calls require iOS 10 or later, which we can't enforce as of yet */
-        if (_errorCallback && _mediaPlayer.state == VLCMediaPlayerStateError &&  !_sessionWillRestart)
-            [[UIApplication sharedApplication] openURL:_errorCallback];
-        else if (_successCallback && !_sessionWillRestart)
-            [[UIApplication sharedApplication] openURL:_successCallback];
-#pragma clang diagnostic pop
+    if (_playbackCompletion) {
+        BOOL finishedPlaybackWithError = _mediaPlayer.state == VLCMediaPlayerStateError &&  !_sessionWillRestart;
+        _playbackCompletion(!finishedPlaybackWithError);
     }
 
     [[self remoteControlService] unsubscribeFromRemoteCommands];
@@ -500,12 +496,12 @@ typedef NS_ENUM(NSUInteger, VLCAspectRatio) {
 
 - (void)setAudioDelay:(float)audioDelay
 {
-    _mediaPlayer.currentAudioPlaybackDelay = 1000000.*audioDelay;
+    _mediaPlayer.currentAudioPlaybackDelay = 1000.*audioDelay;
 }
 
 - (float)audioDelay
 {
-    return _mediaPlayer.currentAudioPlaybackDelay/1000000.;
+    return _mediaPlayer.currentAudioPlaybackDelay/1000.;
 }
 
 - (float)playbackPosition
@@ -520,12 +516,12 @@ typedef NS_ENUM(NSUInteger, VLCAspectRatio) {
 
 - (void)setSubtitleDelay:(float)subtitleDeleay
 {
-    _mediaPlayer.currentVideoSubTitleDelay = 1000000.*subtitleDeleay;
+    _mediaPlayer.currentVideoSubTitleDelay = 1000.*subtitleDeleay;
 }
 
 - (float)subtitleDelay
 {
-    return _mediaPlayer.currentVideoSubTitleDelay/1000000.;
+    return _mediaPlayer.currentVideoSubTitleDelay/1000.;
 }
 
 - (float)hue
@@ -589,19 +585,11 @@ typedef NS_ENUM(NSUInteger, VLCAspectRatio) {
 
 - (void)toggleRepeatMode
 {
-    VLCRepeatMode nextRepeatMode = VLCDoNotRepeat;
-    switch (_listPlayer.repeatMode) {
-        case VLCDoNotRepeat:
-            nextRepeatMode = VLCRepeatCurrentItem;
-            break;
-        case VLCRepeatCurrentItem:
-            nextRepeatMode = VLCRepeatAllItems;
-            break;
-        default:
-            nextRepeatMode = VLCDoNotRepeat;
-            break;
+    if (_listPlayer.repeatMode == VLCRepeatAllItems) {
+        _listPlayer.repeatMode = VLCDoNotRepeat;
+    } else {
+        _listPlayer.repeatMode += 1;
     }
-    _listPlayer.repeatMode = nextRepeatMode;
 }
 
 - (NSInteger)indexOfCurrentAudioTrack
@@ -753,16 +741,12 @@ typedef NS_ENUM(NSUInteger, VLCAspectRatio) {
         case VLCMediaPlayerStateStopped: {
             [_listPlayer.mediaList lock];
             NSUInteger listCount = _listPlayer.mediaList.count;
+            [_listPlayer.mediaList unlock];
             if ([_listPlayer.mediaList indexOfMedia:_mediaPlayer.media] == listCount - 1 && self.repeatMode == VLCDoNotRepeat) {
-                [_listPlayer.mediaList unlock];
                 _sessionWillRestart = NO;
                 [self stopPlayback];
                 return;
-            } else if (listCount > 1) {
-                [_listPlayer.mediaList unlock];
-                [_listPlayer next];
-            } else
-                [_listPlayer.mediaList unlock];
+            }
         } break;
         case VLCMediaPlayerStateESAdded: {
 #if TARGET_OS_IOS
@@ -865,20 +849,44 @@ typedef NS_ENUM(NSUInteger, VLCAspectRatio) {
     [_mediaPlayer jumpBackward:interval];
 }
 
-- (NSString *)screenAspectRatio
+- (UIScreen *)currentScreen
 {
-    UIScreen *screen = [[UIDevice currentDevice] VLCHasExternalDisplay] ? [UIScreen screens][1] : [UIScreen mainScreen];
-    return [NSString stringWithFormat:@"%d:%d", (int)screen.bounds.size.width, (int)screen.bounds.size.height];
+    return [[UIDevice currentDevice] VLCHasExternalDisplay] ? [UIScreen screens][1] : [UIScreen mainScreen];
 }
 
-- (void)switchIPhoneXFullScreen
+- (void)switchToFillToScreen
 {
-    BOOL inFullScreen = _mediaPlayer.videoCropGeometry && [[NSString stringWithUTF8String:_mediaPlayer.videoCropGeometry] isEqualToString:[self screenAspectRatio]];
-    if (inFullScreen) {
+    UIScreen *screen = [self currentScreen];
+    CGSize screenSize = screen.bounds.size;
+
+    CGSize videoSize = _mediaPlayer.videoSize;
+
+    CGFloat ar = videoSize.width / (float)videoSize.height;
+    CGFloat dar = screenSize.width / (float)screenSize.height;
+
+    CGFloat scale;
+
+    if (dar >= ar) {
+        scale = screenSize.width / (float)videoSize.width;
+    } else {
+        scale = screenSize.height / (float)videoSize.height;
+    }
+
+    // Multiplied by screen.scale in consideration of pt to px
+    _mediaPlayer.scaleFactor = scale * screen.scale;
+    _isInFillToScreen = YES;
+}
+
+- (void)toggleFullScreen
+{
+    if (_isInFillToScreen) {
         const char *previousAspectRatio = _currentAspectRatio == VLCAspectRatioDefault ? NULL : [[self stringForAspectRatio:_currentAspectRatio] UTF8String];
         _mediaPlayer.videoAspectRatio = (char *)previousAspectRatio;
+        _mediaPlayer.scaleFactor = 0;
+        _isInFillToScreen = NO;
+    } else {
+        [self switchToFillToScreen];
     }
-    _mediaPlayer.videoCropGeometry = inFullScreen ? NULL : (char *)[[self screenAspectRatio] UTF8String];
 }
 
 - (void)switchAspectRatio
@@ -886,15 +894,20 @@ typedef NS_ENUM(NSUInteger, VLCAspectRatio) {
     _currentAspectRatio = _currentAspectRatio == VLCAspectRatioSixteenToTen ? VLCAspectRatioDefault : _currentAspectRatio + 1;
     switch (_currentAspectRatio) {
         case VLCAspectRatioDefault:
+            _mediaPlayer.scaleFactor = 0;
             _mediaPlayer.videoAspectRatio = NULL;
             _mediaPlayer.videoCropGeometry = NULL;
             break;
         case VLCAspectRatioFillToScreen:
-            _mediaPlayer.videoCropGeometry = (char *)[[self screenAspectRatio] UTF8String];
+            // Reset aspect ratio only with aspectRatio button since we want to keep
+            // the user ratio with double tap.
+            _mediaPlayer.videoAspectRatio = NULL;
+            [self switchToFillToScreen];
             break;
         case VLCAspectRatioFourToThree:
         case VLCAspectRatioSixteenToTen:
         case VLCAspectRatioSixteenToNine:
+            _mediaPlayer.scaleFactor = 0;
             _mediaPlayer.videoCropGeometry = NULL;
             _mediaPlayer.videoAspectRatio = (char *)[[self stringForAspectRatio:_currentAspectRatio] UTF8String];
     }
@@ -965,9 +978,44 @@ typedef NS_ENUM(NSUInteger, VLCAspectRatio) {
 #if !TARGET_OS_TV
 - (BOOL)updateViewpoint:(CGFloat)yaw pitch:(CGFloat)pitch roll:(CGFloat)roll fov:(CGFloat)fov absolute:(BOOL)absolute
 {
+    //adjusting the values
+    if (fabs(yaw) > 180) {
+        yaw = yaw > 0 ? yaw - 360 : yaw + 360;
+    }
+    if (fabs(roll) > 180) {
+        roll = roll > 0 ? roll - 360 : roll + 360;
+    }
+    if (fabs(pitch) > 90) {
+        pitch = pitch > 0 ? pitch - 180 : pitch + 180;
+    }
     return [_mediaPlayer updateViewpoint:yaw pitch:pitch roll:roll fov:fov absolute:absolute];
 }
 
+- (CGFloat)yaw
+{
+    return _mediaPlayer.yaw;
+}
+
+- (CGFloat)pitch
+{
+    return _mediaPlayer.pitch;
+}
+
+- (CGFloat)roll
+{
+    return _mediaPlayer.roll;
+}
+
+- (CGFloat)fov
+{
+    return _mediaPlayer.fov;
+}
+
+- (BOOL)currentMediaIs360Video
+{
+    return [self currentMediaProjection] == VLCMediaProjectionEquiRectangular;
+}
+
 - (NSInteger)currentMediaProjection
 {
     VLCMedia *media = [_mediaPlayer media];
@@ -1048,11 +1096,13 @@ typedef NS_ENUM(NSUInteger, VLCAspectRatio) {
     }
 }
 
-- (BOOL)areHeadphonesPlugged
+- (BOOL)isExternalAudioPlaybackDevicePlugged
 {
+    /* check what output device is currently connected
+     * this code assumes that everything which is not a builtin speaker, must be external */
     NSArray *outputs = [[AVAudioSession sharedInstance] currentRoute].outputs;
-    NSString *portName = [[outputs firstObject] portName];
-    return [portName isEqualToString:@"Headphones"];
+    AVAudioSessionPortDescription *outputDescription = outputs.firstObject;
+    return ![outputDescription.portType isEqualToString:AVAudioSessionPortBuiltInSpeaker];
 }
 
 - (void)audioSessionRouteChange:(NSNotification *)notification
@@ -1063,16 +1113,17 @@ typedef NS_ENUM(NSUInteger, VLCAspectRatio) {
     if (routeChangeReason == AVAudioSessionRouteChangeReasonRouteConfigurationChange)
         return;
 
-    BOOL headphonesPlugged = [self areHeadphonesPlugged];
+    BOOL externalAudioPlaybackDevicePlugged = [self isExternalAudioPlaybackDevicePlugged];
 
-    if (_headphonesWasPlugged && !headphonesPlugged && [_mediaPlayer isPlaying]) {
+    if (_externalAudioPlaybackDevicePlugged && !externalAudioPlaybackDevicePlugged && [_mediaPlayer isPlaying]) {
+        APLog(@"Pausing playback as previously plugged external audio playback device was removed");
         [_mediaPlayer pause];
 #if TARGET_OS_IOS
         [self _savePlaybackState];
 #endif
         [[NSNotificationCenter defaultCenter] postNotificationName:VLCPlaybackControllerPlaybackDidPause object:self];
     }
-    _headphonesWasPlugged = headphonesPlugged;
+    _externalAudioPlaybackDevicePlugged = externalAudioPlaybackDevicePlugged;
 }
 
 #pragma mark - Managing the media item
@@ -1177,6 +1228,11 @@ typedef NS_ENUM(NSUInteger, VLCAspectRatio) {
     _sleepTimer = [NSTimer scheduledTimerWithTimeInterval:timeInterval target:self selector:@selector(stopPlayback) userInfo:nil repeats:NO];
 }
 
+- (BOOL)isPlayingOnExternalScreen
+{
+    return (_renderer || [[UIDevice currentDevice] VLCHasExternalDisplay]);
+}
+
 #pragma mark - background interaction
 
 - (void)applicationWillResignActive:(NSNotification *)aNotification
@@ -1184,8 +1240,8 @@ typedef NS_ENUM(NSUInteger, VLCAspectRatio) {
 #if TARGET_OS_IOS
     [self _savePlaybackState];
 #endif
-
-    if (![[[NSUserDefaults standardUserDefaults] objectForKey:kVLCSettingContinueAudioInBackgroundKey] boolValue]) {
+    if (![self isPlayingOnExternalScreen]
+        && ![[[NSUserDefaults standardUserDefaults] objectForKey:kVLCSettingContinueAudioInBackgroundKey] boolValue]) {
         if ([_mediaPlayer isPlaying]) {
             [_mediaPlayer pause];
             _shouldResumePlaying = YES;
@@ -1197,7 +1253,7 @@ typedef NS_ENUM(NSUInteger, VLCAspectRatio) {
 {
     _preBackgroundWrapperView = _videoOutputViewWrapper;
 
-    if (_mediaPlayer.audioTrackIndexes.count > 0)
+    if (!_renderer && _mediaPlayer.audioTrackIndexes.count > 0)
         [self setVideoTrackEnabled:false];
 }
 
@@ -1288,4 +1344,10 @@ typedef NS_ENUM(NSUInteger, VLCAspectRatio) {
               kVLCSettingHardwareDecoding : [defaults objectForKey:kVLCSettingHardwareDecoding]};
 }
 
+#pragma mark - Renderer
+- (void)setRenderer:(VLCRendererItem * __nullable)renderer
+{
+    _renderer = renderer;
+    [_mediaPlayer setRendererItem:_renderer];
+}
 @end
